# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to axis calculation using sinogram
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/10/2019"


import numpy
import logging
from typing import Optional

from silx.gui import qt
from silx.utils.deprecation import deprecated

from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.settings import EDITING_BACKGROUND_COLOR
from tomwer.synctools.axis import QAxisRP

from ...utils.scandescription import ScanNameLabelAndShape
from .radioaxis import RadioAxisWindow

_logger = logging.getLogger(__file__)


class _AxisTypeSelection(qt.QGroupBox):
    sigSelectionChanged = qt.Signal(str)
    """Signal emitted when the selection changed. Value can be `sinogram` or
    `radio`"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent=parent)
        self.setTitle("compute center of rotation from")
        self.setLayout(qt.QHBoxLayout())
        self._radioRB = qt.QRadioButton("radios", parent=self)
        self.layout().addWidget(self._radioRB)

        self._sinogramRB = qt.QRadioButton("sinogram", parent=self)
        self.layout().addWidget(self._sinogramRB)

        # Signal / Slot connections
        self._radioRB.toggled.connect(self._selectionChanged)
        self._sinogramRB.toggled.connect(self._selectionChanged)

    def getSelection(self):
        if self._radioRB.isChecked():
            return "radio"
        else:
            return "sinogram"

    def setSelection(self, selection):
        if selection == "radio":
            self._radioRB.setChecked(True)
        elif selection == "sinogram":
            self._sinogramRB.setChecked(True)
        else:
            raise ValueError("invalid selection given")

    def _selectionChanged(self, *args, **kwargs):
        self.sigSelectionChanged.emit(self.getSelection())


class AxisWindow(qt.QMainWindow):
    """Main widget for the axis calculation"""

    sigComputationRequested = qt.Signal()
    """signal emitted when a computation is requested"""

    sigApply = qt.Signal()
    """signal emitted when the axis reconstruction parameters are validated"""

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigLockModeChanged = qt.Signal()
    """Signal emitted when the lock on the method change"""

    def __init__(self, axis_params, parent=None, backend=None):
        super().__init__(parent=parent)
        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QVBoxLayout())
        self.setCentralWidget(self._mainWidget)

        self._axis_params = axis_params

        # add scan name
        self._scan_label = ScanNameLabelAndShape(parent=self)
        self._mainWidget.layout().addWidget(self._scan_label)

        # add selection
        self._selectionGB = _AxisTypeSelection(parent=self)
        self._mainWidget.layout().addWidget(self._selectionGB)
        self._selectionGB.hide()

        # add widget for radio and sinogram axis
        self._axisWidget = _AxisWidget(
            parent=self, axis_params=axis_params, backend=backend
        )
        self._mainWidget.layout().addWidget(self._axisWidget)

        self.setCentralWidget(self._mainWidget)

        # set up configuration
        if axis_params.use_sinogram:
            selection = "sinogram"
        else:
            selection = "radio"
        self._selectionGB.setSelection(selection=selection)
        self._selectionGB._selectionChanged(self._selectionGB.getSelection())

        # connect signal / slots
        self._axisWidget.sigValidateRequest.connect(self.sigApply)
        self._axisWidget.sigComputationRequested.connect(self._computationRequested)
        self._axisWidget.sigLockModeChanged.connect(self._lockModeChanged)
        self._selectionGB.sigSelectionChanged.connect(self._axisWidget._axisTypeChanged)
        self._axisWidget.sigSinogramReady.connect(self._computationReady)

        # expose API
        self.setSelection = self._selectionGB.setSelection
        self.getSelection = self._selectionGB.getSelection
        self.hideLockButton = self._axisWidget.hideLockButton
        self.hideApplyButton = self._axisWidget.hideApplyButton
        self.setReconsParams = self._axisWidget.setReconsParams
        self.getPlotWidget = self._axisWidget.getPlotWidget
        self.setMode = self._axisWidget.setMode
        self.manual_uses_full_image = self._axisWidget.manual_uses_full_image
        self.setModeLock = self._axisWidget.setModeLock
        self.getAxis = self._axisWidget.getAxis
        self.isModeLock = self._axisWidget.isModeLock
        self.isValueLock = self._axisWidget.isValueLock
        self.setValueLock = self._axisWidget.setValueLock
        self._setValueLockFrmSettings = self._axisWidget._setValueLockFrmSettings
        self._setModeLockFrmSettings = self._axisWidget._setModeLockFrmSettings
        self.getRadioMode = self._axisWidget.getRadioMode
        self.getEstimatedCor = self._axisWidget.getEstimatedCor
        self.setEstimatedCor = self._axisWidget.setEstimatedCor
        self.sigMethodChanged = self._axisWidget.sigModeChanged

    def setAutoUpdateEstimatedCor(self, value: bool):
        assert isinstance(value, bool)
        self._axisWidget.setAutoUpdateEstimatedCor(value)

    def _disableForProcessing(self, *args, **kwargs):
        self._mainWidget.setEnabled(False)
        self._axisWidget.setEnabled(False)

    def _enableForProcessing(self, *args, **kwargs):
        self._mainWidget.setEnabled(True)
        self._axisWidget.setEnabled(True)

    def _lockModeChanged(self):
        self.sigLockModeChanged.emit()

    def setScan(self, scan, set_position=True):
        """
        set the gui for this scan

        :param TomoBase scan:
        """
        self._scan_label.setScan(scan=scan)
        self._axisWidget.setScan(scan=scan, set_position=set_position)
        self._enableSinogramOpt(scan.scan_range == 360)

    def _enableSinogramOpt(self, b):
        if self._selectionGB.getSelection() == "sinogram" and not b:
            change_selection_to_radio = True
        else:
            change_selection_to_radio = False
        self._selectionGB._sinogramRB.setEnabled(b)
        if change_selection_to_radio:
            self._selectionGB.setSelection("radio")

    def _computationRequested(self):
        self._computationReady()

    def _computationReady(self):
        self.sigComputationRequested.emit()

    def setPosition(self, frm: float, value: float):
        self._axisWidget.setPosition(relative_value=value)


class _AxisWidget(qt.QMainWindow):
    """
    Control which plot and which settings widget is displayed
    """

    sigComputationRequested = qt.Signal()

    sigValidateRequest = qt.Signal()

    sigLockCORValueChanged = qt.Signal(bool)
    """bool: True if locked"""

    sigLockModeChanged = qt.Signal()
    """Signal emitted when the lock mode on the mode change"""

    sigModeChanged = qt.Signal(str)

    sigSinogramReady = qt.Signal()

    def __init__(self, parent, axis_params, backend=None):
        super().__init__(parent)
        self._settingsWidget = None
        self.setDockOptions(qt.QMainWindow.AnimatedDocks)

        self._settingsDockWidget = qt.QDockWidget(parent=self)
        self._settingsDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._settingsDockWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._settingsDockWidget.layout().setSpacing(0)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._settingsDockWidget)

        self._controlWidget = _ControlWidget(parent=self)
        self._controlWidget.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self._controlDockWidget = qt.QDockWidget(parent=self)
        self._controlDockWidget.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self._controlDockWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._controlDockWidget.layout().setSpacing(0)
        self._controlDockWidget.setMaximumHeight(150)
        self._controlDockWidget.setWidget(self._controlWidget)
        self._controlDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._controlDockWidget)

        self._axis_params = axis_params
        self._radioAxis = RadioAxisWindow(
            parent=self, axis=axis_params, backend=backend
        )
        self._plotWidgets = qt.QWidget(self)
        self._plotWidgets.setLayout(qt.QVBoxLayout())
        self.setCentralWidget(self._plotWidgets)
        self._radioAxis.removeDockWidget(self._radioAxis.getSettingsWidgetDocker())

        # expose API
        self.getRadioMode = self._radioAxis.getMode
        self.getEstimatedCor = self._radioAxis.getEstimatedCor
        self.setEstimatedCor = self._radioAxis.setEstimatedCor

        # set up
        if self._axis_params.mode is AxisMode.manual:
            self._controlWidget.setPosition(
                self._axis_params.relative_cor_value,
                self._axis_params.absolute_cor_value,
            )
        self._controlWidget._positionInfo.setAxis(self._axis_params)

        # connect signal / slots
        self._controlWidget.sigComputationRequest.connect(self.sigComputationRequested)
        self._controlWidget.sigValidateRequest.connect(self.sigValidateRequest)
        self._controlWidget.sigLockCORValueChanged.connect(self._CORValueLocked)
        self._radioAxis.sigLockModeChanged.connect(self.sigLockModeChanged)
        self._radioAxis.sigPositionChanged.connect(self._setPositionFrmTuple)
        self._controlWidget._positionInfo.sigRelativeValueSet.connect(
            self._forceRelativePosition
        )
        self._controlWidget._positionInfo.sigAbsolueValueSet.connect(
            self._forceAbsolutePosition
        )

        # set up
        self.addPlot(self._radioAxis)
        self.setActiveWidget(self._radioAxis)

    def _forceRelativePosition(self, value: float):
        self.setMode(AxisMode.manual)
        self.setPosition(
            relative_value=value,
            absolute_value=None,
        )

    def _forceAbsolutePosition(self, value: float):
        self.setMode(AxisMode.manual)
        self.setPosition(
            relative_value=None,
            absolute_value=value,
        )

    def setAutoUpdateEstimatedCor(self, value):
        self._radioAxis.setUpdateAutomaticallyEstimatedCor(value)

    def getAutoUpdateEstimatedCor(self):
        return self._radioAxis.updateAutomaticallyEstimatedCor()

    @deprecated(replacement="getRadioMode")
    def getMode(self):
        return self._radioAxis.getMode()

    def manual_uses_full_image(self, value):
        self._radioAxis.manual_uses_full_image(value)

    def _modeChanged(self):
        self.getAxis().mode = self.getMode()
        self.getAxis().use_sinogram = self.useSinogram()

    def setMode(self, mode):
        mode = AxisMode.from_value(mode)
        self.setActiveWidget(self._radioAxis)
        self._radioAxis.setMode(mode)
        self._axis_params.mode = mode
        self.sigModeChanged.emit(mode.value)

    def _CORValueLocked(self, lock):
        if lock:
            self.setMode(AxisMode.manual)
        self.setModeLock(lock)
        self.sigLockCORValueChanged.emit(lock)

    def setActiveWidget(self, widget):
        for iPlotWidget in range(self._plotWidgets.layout().count()):
            plotWidget = self._plotWidgets.layout().itemAt(iPlotWidget).widget()
            plotWidget.setVisible(widget == plotWidget)
        self.setSettingsWidget(widget.getSettingsWidget())

    def addPlot(self, plot):
        self._plotWidgets.layout().addWidget(plot)

    def setPlotWidget(self, plot):
        self.setCentralWidget(plot)

    def getPlotWidget(self):
        return self.centralWidget()

    def setSettingsWidget(self, widget):
        self._settingsDockWidget.setWidget(widget)

    def _setPositionFrmTuple(self, value):
        self.setPosition(relative_value=value[0])

    def setPosition(self, relative_value: float, absolute_value: float = None) -> None:
        """

        :param float value_ref_tomwer: center of rotation value on tomwer
                                       reference
        :param float value_ref_nabu: center of rotation value on the nabu
                                     reference
        :raises: ValueError if the frm parameter is not recognized
        """
        if (
            absolute_value is None
            and self._axis_params.frame_width is not None
            and relative_value is not None
        ):
            try:
                absolute_value = relative_value + self._axis_params.frame_width / 2.0
            except TypeError:
                absolute_value = None
        if (
            relative_value is None
            and self._axis_params.frame_width is not None
            and absolute_value is not None
        ):
            try:
                relative_value = absolute_value - self._axis_params.frame_width / 2.0
            except TypeError:
                relative_value = None
        self._controlWidget.setPosition(
            value_ref_tomwer=relative_value, value_ref_nabu=absolute_value
        )
        if relative_value is not None:
            self._radioAxis.setXShift(relative_value)
        else:
            self._radioAxis.resetShift()

    def getAxis(self):
        return self._axis_params

    def _axisTypeChanged(self, selection):
        self._axis_params.use_sinogram = False
        self.setActiveWidget(self._radioAxis)

    def setScan(self, scan: TomwerScanBase, set_position):
        """
        set the gui for this scan

        :param TomoBase scan:
        """
        self._radioAxis.setScan(scan=scan)
        if set_position is True and scan.axis_params is not None:
            self.setPosition(
                scan.axis_params.relative_cor_value, scan.axis_params.relative_cor_value
            )
        elif set_position is False:
            if scan.dim_1 is not None:
                # if absolute position is know and relative position not (in the case the user already define absolute position only):
                has_absolute_val_already = not self._controlWidget._positionInfo._absolutePositionQLE.text().startswith(
                    (".", "?")
                )
                has_relative_val_already = not self._controlWidget._positionInfo._relativePositionQLE.text().startswith(
                    (".", "?")
                )
                if (
                    self.getMode() is AxisMode.manual
                    and has_absolute_val_already
                    and not has_relative_val_already
                ):
                    self._controlWidget._updateRelativePosition(width=scan.dim_1)
                else:
                    self._controlWidget._updateAbsolutePosition(width=scan.dim_1)

    def _computationRequested(self) -> None:
        self.sigComputationRequested.emit()

    def hideLockButton(self) -> None:
        self._controlWidget.hideLockButton()

    def hideApplyButton(self) -> None:
        self._controlWidget.hideApplyButton()

    def _setModeLockFrmSettings(self, lock):
        old = self.blockSignals(True)
        self._radioAxis._setModeLockFrmSettings(lock)
        self.blockSignals(old)

    def _setValueLockFrmSettings(self, lock):
        old = self.blockSignals(True)
        self.setValueLock(lock)
        self.blockSignals(old)

    def setModeLock(self, lock):
        assert type(lock) is bool
        self._radioAxis.setLocked(lock)

    def isModeLock(self):
        return self._radioAxis.isModeLock()

    def isValueLock(self):
        return self._controlWidget.isValueLock()

    def setValueLock(self, lock):
        self._controlWidget.setValueLock(lock)

    def setReconsParams(self, recons_params):
        self._axis_params = recons_params
        self._radioAxis.setReconsParams(axis=recons_params)
        if recons_params.mode is AxisMode.manual:
            self._controlWidget.setPosition(
                self._axis_params.relative_cor_value,
                self._axis_params.absolute_cor_value,
            )


class _ControlWidget(qt.QWidget):
    """
    Widget to lock cor position or compute it or validate it and to
    display the cor value
    """

    sigComputationRequest = qt.Signal()
    """Signal emitted when user request a computation from the settings"""

    sigValidateRequest = qt.Signal()
    """Signal emitted when user validate the current settings"""

    sigLockCORValueChanged = qt.Signal(bool)
    """Signal emitted when the user lock the cor value. Param: True if lock"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        # display 'center' information
        self._positionInfo = _PositionInfoWidget(parent=self)
        self.layout().addWidget(self._positionInfo)

        self._buttons = qt.QWidget(parent=self)
        self._buttons.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._buttons)

        self._lockBut = PadlockButton(parent=self)
        self._lockBut.setAutoDefault(False)
        self._lockBut.setDefault(False)

        self._buttons.layout().addWidget(self._lockBut)
        self._lockLabel = qt.QLabel("lock cor value", parent=self)
        self._buttons.layout().addWidget(self._lockLabel)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._buttons.layout().addWidget(spacer)

        self._computeBut = qt.QPushButton("compute", parent=self)
        self._buttons.layout().addWidget(self._computeBut)
        style = qt.QApplication.style()
        applyIcon = style.standardIcon(qt.QStyle.SP_DialogApplyButton)
        self._applyBut = qt.QPushButton(applyIcon, "validate", parent=self)
        self._buttons.layout().addWidget(self._applyBut)
        self.layout().addWidget(self._buttons)

        # set up
        self._positionInfo.setPosition(None, None)

        # make connection
        self._computeBut.pressed.connect(self._needComputation)
        self._applyBut.pressed.connect(self._validate)
        self._lockBut.sigLockChanged.connect(self._lockValueChanged)

    def hideLockButton(self) -> None:
        self._lockLabel.hide()
        self._lockBut.hide()

    def hideApplyButton(self) -> None:
        self._applyBut.hide()

    def _lockValueChanged(self):
        self.sigLockCORValueChanged.emit(self._lockBut.isLocked())
        self._computeBut.setEnabled(not self._lockBut.isLocked())

    def _needComputation(self, *arg, **kwargs):
        """callback when the radio line changed"""
        self.sigComputationRequest.emit()

    def _validate(self):
        self.sigValidateRequest.emit()

    def setPosition(self, value_ref_tomwer, value_ref_nabu):
        self._positionInfo.setPosition(
            relative_cor=value_ref_tomwer, abs_cor=value_ref_nabu
        )

    def _updateAbsolutePosition(self, width):
        self._positionInfo._updateAbsolutePosition(width=width)

    def _updateRelativePosition(self, width):
        self._positionInfo._updateRelativePosition(width=width)

    def isValueLock(self):
        return self._lockBut.isLocked()

    def setValueLock(self, lock):
        self._lockBut.setLock(lock)


class _PositionInfoValidator(qt.QDoubleValidator):
    def validate(self, a0: str, a1: int):
        if "..." in a0 or a0 == "?":
            return (qt.QDoubleValidator.Acceptable, a0, a1)
        else:
            return super().validate(a0, a1)


class _PositionInfoWidget(qt.QWidget):
    """Widget used to display information relative to the current position"""

    sigRelativeValueSet = qt.Signal(float)
    """Emit when the user define manually the relative value"""
    sigAbsolueValueSet = qt.Signal(float)
    """Emit when the user define manually the absolute value"""

    def __init__(self, parent, axis=None):
        self._axis = None
        self._defaultBackgroundColor = None
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        centerLabel = qt.QLabel("center", parent=self)
        centerLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        font = centerLabel.font()
        font.setBold(True)
        centerLabel.setFont(font)

        self.layout().addWidget(centerLabel, 0, 0, 1, 1)
        self.layout().addWidget(qt.QLabel(" (relative):"), 0, 1, 1, 1)

        self._relativePositionQLE = qt.QLineEdit("", parent=self)
        self._positionValidator = _PositionInfoValidator(self)
        self._relativePositionQLE.setValidator(self._positionValidator)
        self._relativePositionQLE.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self._relativePositionQLE.setStyleSheet("color: red")
        self.layout().addWidget(self._relativePositionQLE, 0, 2, 1, 1)

        centerLabel = qt.QLabel("center", parent=self)
        centerLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        font = centerLabel.font()
        font.setBold(False)
        centerLabel.setFont(font)

        self.layout().addWidget(centerLabel, 1, 0, 1, 1)
        self.layout().addWidget(qt.QLabel(" (absolute):"), 1, 1, 1, 1)
        self._absolutePositionQLE = qt.QLineEdit("", parent=self)
        self._absolutePositionQLE.setValidator(self._positionValidator)
        self._absolutePositionQLE.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self._absolutePositionQLE.setStyleSheet("color: #ff8c00")
        self.layout().addWidget(self._absolutePositionQLE, 1, 2, 1, 1)

        if axis:
            self.setAxis(axis)

        # connect signal / slot
        self._relativePositionQLE.textEdited.connect(
            self._userStartEditingRelativePosition
        )
        self._relativePositionQLE.editingFinished.connect(
            self._userUpdatedRelativePosition
        )
        self._absolutePositionQLE.textEdited.connect(
            self._userStartEditingAbsolutePosition
        )

        self._absolutePositionQLE.editingFinished.connect(
            self._userUpdatedAbsolutePosition
        )

    def _getDefaultBackgroundColor(self):
        if self._defaultBackgroundColor is None:
            self._defaultBackgroundColor = self.palette().color(
                self._absolutePositionQLE.backgroundRole()
            )
        return self._defaultBackgroundColor

    def _userStartEditingRelativePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._relativePositionQLE.backgroundRole(), EDITING_BACKGROUND_COLOR
        )
        self._relativePositionQLE.setPalette(palette)

    def _userStartEditingAbsolutePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._absolutePositionQLE.backgroundRole(), EDITING_BACKGROUND_COLOR
        )
        self._absolutePositionQLE.setPalette(palette)

    def _userUpdatedRelativePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._relativePositionQLE.backgroundRole(),
            self._getDefaultBackgroundColor(),
        )
        self._relativePositionQLE.setPalette(palette)
        if self._relativePositionQLE.text().startswith((".", "?")):
            return
        else:
            value = float(self._relativePositionQLE.text())
            # make sure we only emit the signal if the value changed (and when the Qline has been edited).
            if self._axis.relative_cor_value is None or (
                self._axis is not None
                and not numpy.isclose(value, self._axis.relative_cor_value, atol=1e-3)
            ):
                self.sigRelativeValueSet.emit(value)

    def _userUpdatedAbsolutePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._absolutePositionQLE.backgroundRole(),
            self._getDefaultBackgroundColor(),
        )
        self._absolutePositionQLE.setPalette(palette)
        if self._absolutePositionQLE.text().startswith((".", "?")):
            return
        else:
            value = float(self._absolutePositionQLE.text())
            # make sure we only emit the signal if the value changed (and when the Qline has been edited).
            if self._axis.absolute_cor_value is None or (
                self._axis is not None
                and not numpy.isclose(value, self._axis.absolute_cor_value, atol=1e-3)
            ):
                self.sigAbsolueValueSet.emit(value)

    def setAxis(self, axis):
        assert isinstance(axis, QAxisRP)
        if axis == self._axis:
            return
        self._axis = axis

    def getPosition(self):
        return float(self._relativePositionQLE.text())

    def setPosition(self, relative_cor: Optional[float], abs_cor: Optional[float]):
        if relative_cor is None:
            self._relativePositionQLE.setText("?")
        elif isinstance(relative_cor, str):
            self._relativePositionQLE.setText(relative_cor)
        else:
            self._relativePositionQLE.setText(f"{relative_cor:.3f}")
        if abs_cor is None:
            self._absolutePositionQLE.setText("?")
        else:
            self._absolutePositionQLE.setText(f"{abs_cor:.3f}")

    def _updateAbsolutePosition(self, width):
        if width is not None:
            try:
                rel_value = float(self._relativePositionQLE.text())
            except Exception:
                return
            else:
                abs_value = float(rel_value) + width / 2.0
                self._absolutePositionQLE.setText(str(abs_value))

    def _updateRelativePosition(self, width):
        if width is not None:
            try:
                abs_value = float(self._absolutePositionQLE.text())
            except Exception:
                return
            else:
                rel_value = float(abs_value) - width / 2.0
                self._relativePositionQLE.setText(str(rel_value))



from silx.gui import qt
from typing import Optional


class CreateRawDataScreenshotsWidget(qt.QWidget):
    """
    Widget to allow the user define the screenshot to make of the raw data
    """

    sigConfigChanged = qt.Signal()
    """emit when the configuration change"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())
        # projections
        self._projectionsCB = qt.QCheckBox("raw projections each", self)
        self._projectionsSB = qt.QSpinBox(parent)
        self._projectionsSB.setSuffix("°")
        self.layout().addRow(self._projectionsCB, self._projectionsSB)
        # flat field
        self._flatFieldCB = qt.QCheckBox("first raw flat")
        self.layout().addRow(self._flatFieldCB)
        # dark field
        self._darkFieldCB = qt.QCheckBox("first raw dark")
        self.layout().addRow(self._darkFieldCB)

        # set up
        self._projectionsCB.setChecked(True)
        self._flatFieldCB.setChecked(True)
        self._darkFieldCB.setChecked(True)

        # connect signal / slot
        self._projectionsCB.toggled.connect(self._changed)
        self._projectionsSB.valueChanged.connect(self._changed)
        self._projectionsSB.setRange(0, 360)
        self._projectionsSB.setValue(90)
        self._flatFieldCB.toggled.connect(self._changed)
        self._darkFieldCB.toggled.connect(self._changed)

    def _changed(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        return {
            "raw_projections_required": self._projectionsCB.isChecked(),
            "raw_projections_each": self._projectionsSB.value(),
            "raw_darks_required": self._darkFieldCB.isChecked(),
            "raw_flats_required": self._flatFieldCB.isChecked(),
        }

    def setRawProjections(self, required: bool, each_proj: Optional[int]):
        self._projectionsCB.setChecked(required)
        if each_proj is not None:
            self._projectionsSB.setValue(int(each_proj))

    def setFlatFieldRequired(self, required: bool):
        self._flatFieldCB.setChecked(required)

    def setDarkFieldRequired(self, required: bool):
        self._darkFieldCB.setChecked(required)

    def setConfiguration(self, configuration: dict):
        assert isinstance(configuration, dict)
        # handle raw projections
        raw_projections_required = configuration.get(
            "raw_projections_required", self._projectionsCB.isChecked()
        )
        raw_projections_each = configuration.get("raw_projections_each", None)
        self.setRawProjections(
            required=raw_projections_required, each_proj=raw_projections_each
        )
        # handle flat field
        flat_field_required = configuration.get("raw_flats_required", None)
        if flat_field_required is not None:
            self.setFlatFieldRequired(flat_field_required)
        # handle flat field
        dark_field_required = configuration.get("raw_darks_required", None)
        if dark_field_required is not None:
            self.setDarkFieldRequired(dark_field_required)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "31/05/2018"


import os
import weakref
from collections import OrderedDict

from silx.gui import qt

from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase


class ScanObservation(qt.QWidget):
    """Widget used for the scans observation"""

    HEADER = ("acquisition", "status", "N projections", "type")

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self._onGoingObservations = None
        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(qt.QLabel(""))

        self.observationTable = qt.QTableView(parent=parent)
        self.observationTable.setSelectionBehavior(qt.QAbstractItemView.SelectRows)
        self.observationTable.setModel(
            _ObservedScanModel(parent=self.observationTable, header=self.HEADER)
        )
        self.observationTable.resizeColumnsToContents()
        self.observationTable.setSortingEnabled(True)
        self.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)

        self.layout().addWidget(self.observationTable)
        header = self.observationTable.horizontalHeader()
        if qt.qVersion() < "5.0":
            setResizeMode = header.setResizeMode
        else:
            setResizeMode = header.setSectionResizeMode
        setResizeMode(0, qt.QHeaderView.Fixed)
        setResizeMode(1, qt.QHeaderView.Stretch)
        setResizeMode(2, qt.QHeaderView.Stretch)

        header.setSectionResizeMode(0, qt.QHeaderView.Interactive)
        header.setStretchLastSection(True)

    @property
    def onGoingObservations(self):
        if self._onGoingObservations:
            return self._onGoingObservations()
        else:
            return None

    def setOnGoingObservations(self, onGoingObservations):
        """
        will update the table to display the observations contained in
        onGoingObservations

        :param onGoingObservations: the obsevations observed to display
        """
        if self.onGoingObservations:
            self.onGoingObservations.sigObsAdded.disconnect(self.addObservation)
            self.onGoingObservations.sigObsRemoved.disconnect(self.removeObservation)
            self.onGoingObservations.sigObsStatusReceived.disconnect(self.update)

        self._onGoingObservations = weakref.ref(onGoingObservations)
        self.onGoingObservations.sigObsAdded.connect(self.addObservation)
        self.onGoingObservations.sigObsRemoved.connect(self.removeObservation)
        self.onGoingObservations.sigObsStatusReceived.connect(self.update)

    def update(self, scan, status):
        """

        :param str scan: the updated scan
        :param str status: the status of the updated scan
        """
        self.observationTable.model().update(scan, status)

    def addObservation(self, scan):
        """

        :param scan: the scan observed
        """
        self.observationTable.model().add(scan, "starting")

    def removeObservation(self, scan):
        """

        :param scan: the scan removed
        """
        self.observationTable.model().remove(scan)

    def clear(self):
        self.observationTable.model().clear()


class _ObservedScanModel(qt.QAbstractTableModel):
    def __init__(self, parent, header, *args):
        qt.QAbstractTableModel.__init__(self, parent, *args)
        self.header = header
        self.observations = OrderedDict()

    def add(self, scan, status):
        self.observations[scan] = status
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def remove(self, scan):
        if scan in self.observations:
            del self.observations[scan]
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def update(self, scan, status):
        self.observations[scan] = status
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def clear(self):
        self.observations = OrderedDict()
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def rowCount(self, parent=None):
        return len(self.observations)

    def columnCount(self, parent=None):
        return len(self.header)

    def sort(self, col, order):
        self.layoutAboutToBeChanged.emit()
        if self.observations is None:
            return

        to_order = {}
        for observation in self.observations.keys():
            to_order[str(observation)] = observation

        ordering = sorted(list(to_order.keys()))
        if order == qt.Qt.DescendingOrder:
            ordering = reversed(ordering)
        _observations = OrderedDict()
        for str_key in ordering:
            key = to_order[str_key]
            _observations[key] = self.observations[key]

        self.observations = _observations
        self.layoutChanged.emit()

    def data(self, index, role):
        if index.isValid() is False:
            return None

        if role not in (qt.Qt.DisplayRole, qt.Qt.ToolTipRole):
            return None

        obs = list(self.observations.keys())[index.row()]
        if index.column() == 0:
            if role == qt.Qt.ToolTipRole:
                return obs
            elif isinstance(obs, (TomwerScanBase, BlissScan)):
                return str(obs)
            else:
                return os.path.basename(obs)
        elif index.column() == 1:
            return self.observations[obs]
        elif index.column() == 2:
            if isinstance(obs, TomwerScanBase):
                return obs.tomo_n or 0
            elif isinstance(obs, BlissScan):
                return f"(at least) {obs.n_acquired or '?'} over {obs.tomo_n or '?'}"
            elif os.path.exists(obs) and os.path.isdir(obs):
                return str(len(os.listdir(obs)))
            else:
                return None
        elif index.column() == 3:
            if isinstance(obs, TomwerScanBase):
                return obs.type
            elif isinstance(obs, BlissScan):
                return "hdf5"
            elif NXtomoScan.directory_contains_scan(directory=obs):
                return "hdf5"
            else:
                return "edf"
        else:
            return None

    def headerData(self, col, orientation, role):
        if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
            return self.header[col]
        return None



from tomwer.core.scan.nxtomoscan import NXtomoScanIdentifier, NXtomoScan
from tomwer.core.utils.deprecation import deprecated_warning


class HDF5TomoScanIdentifier(NXtomoScanIdentifier):
    def __init__(self, metadata=None, *args, **kwargs):
        deprecated_warning(
            name="tomwer.core.scan.nxtomoscan.NXtomoScanIdentifier",
            type_="class",
            reason="improve readibility",
            since_version="1.3",
            replacement="VolumeSymbolicLinkTask",
        )
        super().__init__(metadata, *args, **kwargs)


class HDF5TomoScan(NXtomoScan):
    def __init__(self, scan, entry, index=None, overwrite_proc_file=False):
        deprecated_warning(
            name="tomwer.core.scan.nxtomoscan.NXtomoScan",
            type_="class",
            reason="improve readibility",
            since_version="1.3",
            replacement="VolumeSymbolicLinkTask",
        )
        super().__init__(scan, entry, index, overwrite_proc_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import signal
import sys
import time
from typing import Union

import silx
from silx.gui import qt
from silx.io.utils import h5py_read_dataset

from tomwer.core.process import utils as core_utils
from tomwer.core.process.reconstruction.axis.axis import AxisTask, NoAxisUrl
from tomwer.core.process.reconstruction.darkref.darkrefs import (
    requires_reduced_dark_and_flat,
)
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisTask
from tomwer.core.process.task import Task
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.reconstruction.saaxis.saaxis import SAAxisWindow as _SAAxisWindow
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.saaxis import QSAAxisParams

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class SAAxisThread(qt.QThread):
    """
    Thread to call nabu and reconstruct one slice with several cor value
    """

    def init(self, data, configuration, dump_roi):
        self.scan = data
        self._configuration = configuration
        self._dump_roi = dump_roi

    def run(self) -> None:
        process = SAAxisTask(
            process_id=None,
            inputs={
                "data": self.scan,
                "sa_axis_params": self._configuration,
                "serialize_output_data": False,
            },
        )
        process.dump_roi = self._dump_roi
        t0 = time.time()
        process.run()
        print(f"execution time is {time.time() - t0}")


class SAAxisWindow(_SAAxisWindow):
    def __init__(self, parent=None, dump_roi=False, backend=None):
        self._scan = None
        super().__init__(parent, backend=backend)
        # thread for computing cors
        self._processingThread = SAAxisThread()
        self._processingThread.finished.connect(self._threadedProcessEnded)
        self.sigStartSinogramLoad.connect(self._callbackStartLoadSinogram)
        self.sigEndSinogramLoad.connect(self._callbackEndLoadSinogram)
        self._dump_roi = dump_roi

        # hide the validate button
        self._saaxisControl._applyBut.hide()
        self.hideAutoFocusButton()

    def _launchReconstructions(self):
        if self._processingThread.isRunning():
            _logger.error(
                "a calculation is already launch. You must wait for "
                "it to end prior to launch a new one"
            )
        else:
            qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
            self._processingThread.init(
                configuration=self.getConfiguration(),
                data=self.getScan(),
                dump_roi=self._dump_roi,
            )
            self._processingThread.start()

    def _threadedProcessEnded(self):
        saaxis_params = self._processingThread.scan.saaxis_params
        if saaxis_params is None:
            scores = None
        else:
            scores = saaxis_params.scores
        scan = self.getScan()
        assert scan is not None, "scan should have been set"
        self.setCorScores(
            scores, img_width=scan.dim_1, score_method=self.getScoreMethod()
        )
        if scan.saaxis_params.autofocus is not None:
            self.setCurrentCorValue(scan.saaxis_params.autofocus)
        self.showResults()
        qt.QApplication.restoreOverrideCursor()

    def _callbackStartLoadSinogram(self):
        print(f"start loading sinogram for {self.getScan()}. Can take some time")

    def _callbackEndLoadSinogram(self):
        print(f"sinogram loaded for {self.getScan()} loaded.")

    def close(self) -> None:
        self._stopProcessingThread()
        super().close()

    def _stopProcessingThread(self):
        if self._processingThread:
            self._processingThread.terminate()
            self._processingThread.wait(500)
            self._processingThread = None

    def stop(self):
        self._stopProcessingThread()

    def _computeEstimatedCor(self) -> Union[float, None]:
        scan = self.getScan()
        if scan is None:
            return
        _cor_estimation_process = AxisTask(
            inputs={
                "axis_params": self.getQAxisRP(),
                "data": scan,
                "serialize_output_data": False,
            }
        )

        _logger.info(f"{scan} - start cor estimation for")
        qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
        try:
            _cor_estimation_process.compute(scan=scan, wait=True)
        except NoAxisUrl:
            qt.QApplication.restoreOverrideCursor()
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = (
                "Unable to find url to compute the axis, please select them "
                "from the `axis input` tab"
            )
            msg.setText(text)
            msg.exec_()
            return None
        else:
            self.setEstimatedCorPosition(
                value=scan.axis_params.relative_cor_value,
            )
            qt.QApplication.restoreOverrideCursor()
            self.getAutomaticCorWindow().hide()
            return scan.axis_params.relative_cor_value

    def setDumpScoreROI(self, dump):
        self._dump_score_roi = dump


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus "
        "provide the master file",
        default=None,
    )
    parser.add_argument(
        "--entry", help="For Nexus files: entry in the master file", default=None
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--read-existing",
        dest="read_existing",
        action="store_true",
        default=False,
        help="Load latest sa-delta-beta processing from *_tomwer_processes.h5 "
        "if exists",
    )
    parser.add_argument(
        "--dump-roi",
        dest="dump_roi",
        action="store_true",
        default=False,
        help="Save roi where the score is computed on the .hdf5",
    )
    parser.add_argument(
        "--use-opengl-plot",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)
    # define the process_index is any tomwer_processes_existing
    if options.debug:
        _logger.setLevel(logging.DEBUG)

    scan = ScanFactory.create_scan_object(
        scan_path=options.scan_path, entry=options.entry
    )
    requires_reduced_dark_and_flat(scan=scan, logger_=_logger)

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    window = SAAxisWindow(dump_roi=options.dump_roi)
    window.setWindowTitle("saaxis")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    if scan.axis_params is None:
        scan.axis_params = QAxisRP()
    if scan.saaxis_params is None:
        scan.saaxis_params = QSAAxisParams()
    # force load of reduced_flats and reduced_darks
    scan.reduced_flats
    scan.reduced_darks
    window.setScan(scan)
    window.setDumpScoreROI(options.dump_roi)
    if options.read_existing is True:
        scores, selected = _load_latest_scores(scan)
        if scores is not None:
            window.setCorScores(scores, score_method="standard deviation")
            if selected not in (None, "-"):
                window.setCurrentCorValue(selected)

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.aboutToQuit.connect(window.stop)
    exit(app.exec_())


def _load_latest_scores(scan) -> tuple:
    """

    :param scan:
    :return: loaded score as (scores, selected). Scores is None or a dict.
             selected is None or a float
    """
    scores = None
    selected = None
    if scan.process_file is None:
        _logger.warning(
            "Unable to find process file. Unable to read " "existing processing"
        )
        return scores, selected

    with EntryReader(scan.process_file_url) as h5f:
        latest_saaxis_node = Task.get_most_recent_process(h5f, SAAxisTask)
        if latest_saaxis_node and "results" in latest_saaxis_node:
            scores = core_utils.get_scores(latest_saaxis_node)
            if "center_of_rotation" in latest_saaxis_node["results"]:
                selected = h5py_read_dataset(
                    latest_saaxis_node["results"]["center_of_rotation"]
                )
        else:
            _logger.warning(f"no results found for {scan}")
    return scores, selected


def getinputinfo():
    return "tomwer saaxis [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"

import logging
import os
import shutil
import tempfile
import time
import weakref

import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomoscan.esrf.volume.edfvolume import EDFVolume

from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.stacks import RadioStack, SliceStack
from tomwer.tests.utils import skip_gui_test

logging.disable(logging.INFO)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestSliceStack(TestCaseQt):
    """unit test for the :class:_ImageStack widget"""

    def setUp(self):
        super().setUp()
        self._widget = SliceStack()
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)

    def tearDown(self):
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        super().tearDown()

    def test_hdf5(self):
        self.assertEqual(len(self._widget._scans), 0)
        n_scan = 5
        n_slice_per_scan = 1
        with tempfile.TemporaryDirectory() as root_dir:
            for i_scan in range(n_scan):
                scan = MockNXtomo(
                    scan_path=os.path.join(root_dir, f"scan{i_scan}"),
                    n_proj=10,
                    n_ini_proj=10,
                    scan_range=180,
                    dim=10,
                ).scan
                volume = EDFVolume(
                    folder=os.path.join(scan.path, "edf_volume"),
                    data=numpy.random.random(10 * 10 * n_slice_per_scan).reshape(
                        n_slice_per_scan, 10, 10
                    ),
                )
                volume.save()
                scan.set_latest_reconstructions((volume,))
                assert len(scan.latest_reconstructions) == 1
                self._widget.addLeafScan(scan)
        self.assertEqual(len(self._widget._scans), n_scan)
        self.assertEqual(len(self._widget.extractImages()), n_scan * n_slice_per_scan)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestRadioStack(TestCaseQt):
    """
    Test for the RadioStack
    The main part of test on mode is done on silx side, test of silx.gui.plot.ImageStack
    """

    def setUp(self):
        super().setUp()
        self._widget = RadioStack()
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.spinBox = weakref.ref(self._widget._viewer._qspinbox)
        self.slider = weakref.ref(self._widget._viewer._qslider)
        self._tmp_dir = tempfile.mkdtemp()
        os.makedirs(self._tmp_dir, exist_ok=True)
        self._scan = MockNXtomo(
            os.path.join(self._tmp_dir, "my_scan"),
            n_proj=30,
            dim=10,
        ).scan

    def tearDown(self):
        shutil.rmtree(self._tmp_dir)
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        super().tearDown()

    def _waitImages(self):
        for _ in range(5):
            self.qapp.processEvents()
            time.sleep(0.1)

    def testASAPMode(self):
        self._widget.setLoadingMode("load ASAP")
        self._widget.setForceSync(True)
        self._widget.addLeafScan(self._scan.master_file)
        self._waitImages()

        self.spinBox().setValue(20)
        self._waitImages()

        self.slider().setValue(10)
        self.qapp.processEvents()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to semi-automatic axis calculation
"""


__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "05/02/2021"


from collections.abc import Iterable

from silx.gui import qt
from pyunitsystem import metricsystem


class DimensionWidget(qt.QGroupBox):
    """
    Simple widget to display value over 3 dimensions

    :param parent:
    :param str title: QGroupBox title
    :param dims_name: name of the dimension. If set will be store in each
                      QDoubleLine prefix
    :type dims_name: tuple
    :param dims_colors: color associated to the three dimensions if any
    :type dims_colors: tuple
    """

    valuesChanged = qt.Signal()
    """Signal emitted when a value change"""

    def __init__(
        self, parent=None, title=None, dims_name=None, dims_colors=None, title_size=10
    ):
        qt.QGroupBox.__init__(self, parent)
        self.setFont(qt.QFont("Arial", title_size))
        assert title is not None
        assert dims_name is None or (
            isinstance(dims_name, Iterable) and len(dims_name) == 3
        )
        assert dims_colors is None or (
            isinstance(dims_colors, Iterable) and len(dims_colors) == 3
        )
        self._dim0Value = 1.0 * metricsystem.millimeter.value
        self._dim1Value = 1.0 * metricsystem.millimeter.value
        self._dim2Value = 1.0 * metricsystem.millimeter.value
        self._displayUnit = metricsystem.MetricSystem.MILLIMETER
        # defined unit to display values. Always stored in m (International
        # System)
        ## set GUI
        self.setTitle(title)
        self.setLayout(qt.QHBoxLayout())
        # dim 0
        self._dim0ValueQLE = qt.QDoubleSpinBox(self)
        if dims_name is not None:
            self._dim0ValueQLE.setPrefix(dims_name[0])
        self._dim0ValueQLE.setRange(0, 999999999999)
        self._dim0ValueQLE.setDecimals(10)
        self._dim0ValueQLE.setSingleStep(0.0001)
        self._dim0ValueQLE.setValue(self._getDim0DisplayValue())
        if dims_colors is not None:
            stylesheet = f"background-color: {dims_colors[0]}"
            self._dim0ValueQLE.setStyleSheet(stylesheet)
        self.layout().addWidget(self._dim0ValueQLE)
        # dim 1
        self._dim1ValueQLE = qt.QDoubleSpinBox(self)
        if dims_name is not None:
            self._dim1ValueQLE.setPrefix(dims_name[1])
        self._dim1ValueQLE.setRange(0, 999999999999)
        self._dim1ValueQLE.setDecimals(10)
        self._dim1ValueQLE.setSingleStep(0.0001)
        self._dim1ValueQLE.setValue(self._getDim1DisplayValue())
        if dims_colors is not None:
            stylesheet = f"background-color: {dims_colors[1]}"
            self._dim1ValueQLE.setStyleSheet(stylesheet)
        self.layout().addWidget(self._dim1ValueQLE)
        # dim 2
        self._dim2ValueQLE = qt.QDoubleSpinBox(self)
        if dims_name is not None:
            self._dim2ValueQLE.setPrefix(dims_name[2])
        self._dim2ValueQLE.setRange(0, 999999999999)
        self._dim2ValueQLE.setDecimals(10)
        self._dim2ValueQLE.setSingleStep(0.0001)
        self._dim2ValueQLE.setValue(self._getDim2DisplayValue())
        if dims_colors is not None:
            stylesheet = f"background-color: {dims_colors[2]}"
            self._dim2ValueQLE.setStyleSheet(stylesheet)
        self.layout().addWidget(self._dim2ValueQLE)

        # set up
        self.setUnit(self._displayUnit)

        # connect signal / slot
        self._dim0ValueQLE.editingFinished.connect(self._userSetDim0)
        self._dim1ValueQLE.editingFinished.connect(self._userSetDim1)
        self._dim2ValueQLE.editingFinished.connect(self._userSetDim2)

    def _getDim0DisplayValue(self) -> float:
        return self._dim0Value / self._displayUnit.value

    def _getDim1DisplayValue(self) -> float:
        return self._dim1Value / self._displayUnit.value

    def _getDim2DisplayValue(self) -> float:
        return self._dim2Value / self._displayUnit.value

    def setUnit(self, unit):
        """
        define with which unit we should display the size
        :param unit: metric to be used for display. Internally this is always
                     stored using the international metric system
        """
        self._displayUnit = metricsystem.MetricSystem.from_value(unit)
        for widget in (self._dim0ValueQLE, self._dim1ValueQLE, self._dim2ValueQLE):
            widget.setSuffix(str(self.unit()))
        # update displayed values
        old = self.blockSignals(True)
        self._dim0ValueQLE.setValue(self._getDim0DisplayValue())
        self._dim1ValueQLE.setValue(self._getDim1DisplayValue())
        self._dim2ValueQLE.setValue(self._getDim2DisplayValue())
        self.blockSignals(old)

    def unit(self) -> metricsystem.MetricSystem:
        """

        :return: metric system used for display
        :rtype: metricsystem.MetricSystem
        """
        return self._displayUnit

    def setValues(self, dim0, dim1, dim2, unit="mm") -> None:
        """

        :param float dim0: value to dim0
        :param float dim1: value to dim1
        :param float dim2: value to dim2
        :param Union[str,MetricSystem] unit: unit used for the provided values
        :return:
        """
        old = self.blockSignals(True)
        self.setDim0value(value=dim0, unit=unit)
        self.setDim1value(value=dim1, unit=unit)
        self.setDim2value(value=dim2, unit=unit)
        self.blockSignals(old)
        self.valuesChanged.emit()

    def getValues(self) -> tuple:
        """

        :return: (dim0 value, dim1 value, dim2 value, unit)
        :rtype: tuple
        """
        return (
            self.getDim0Value()[0],
            self.getDim1Value()[0],
            self.getDim2Value()[0],
            metricsystem.MetricSystem.METER,
        )

    def getDim0Value(self) -> tuple:
        """Return Dim 0 value and unit. Unit is always metter"""
        return self._dim0Value, metricsystem.MetricSystem.METER

    def setDim0value(self, value, unit="mm"):
        """

        :param value: value to set to dim 0.
        :param Union[str,MetricSystem] unit:
        :return:
        """
        self._dim0Value = value * metricsystem.MetricSystem.from_value(unit).value
        old = self.blockSignals(True)
        self._dim0ValueQLE.setValue(self._dim0Value)
        self.blockSignals(old)
        self.valuesChanged.emit()

    def getDim1Value(self) -> tuple:
        """Return Dim 0 value and unit. Unit is always metter"""
        return self._dim1Value, metricsystem.MetricSystem.METER

    def setDim1value(self, value, unit="mm"):
        """

        :param value: value to set to dim 1.
        :param Union[str,MetricSystem] unit:
        :return:
        """
        self._dim1Value = value * metricsystem.MetricSystem.from_value(unit).value
        old = self.blockSignals(True)
        self._dim1ValueQLE.setValue(self._dim1Value)
        self.blockSignals(old)
        self.valuesChanged.emit()

    def getDim2Value(self) -> tuple:
        """Return Dim 0 value and unit. Unit is always metter"""
        return self._dim2Value, metricsystem.MetricSystem.METER

    def setDim2value(self, value, unit="mm"):
        """

        :param value: value to set to dim 2.
        :param Union[str,MetricSystem] unit:
        :return:
        """
        self._dim2Value = value * metricsystem.MetricSystem.from_value(unit).value
        old = self.blockSignals(True)
        self._dim2ValueQLE.setValue(self._dim2Value)
        self.blockSignals(old)
        self.valuesChanged.emit()

    def _valuesChanged(self, *args, **kwargs):
        self.valuesChanged.emit()

    def _userSetDim0(self):
        """callback when the user modify the dim 0 QDSP"""
        old = self.blockSignals(True)
        self._dim0Value = self._dim0ValueQLE.value() * self.unit().value
        self.blockSignals(old)
        self.valuesChanged.emit()

    def _userSetDim1(self):
        """callback when the user modify the dim 1 QDSP"""
        old = self.blockSignals(True)
        self._dim1Value = self._dim1ValueQLE.value() * self.unit().value
        self.blockSignals(old)
        self.valuesChanged.emit()

    def _userSetDim2(self):
        """callback when the user modify the dim 2 QDSP"""
        old = self.blockSignals(True)
        self._dim2Value = self._dim2ValueQLE.value() * self.unit().value
        self.blockSignals(old)
        self.valuesChanged.emit()



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to manage observations. Initially on files.
Observations are runned on a thread and run each n seconds.
They are manage by thread and signals
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/02/2017"


import logging
import os
import threading
import time
from collections import OrderedDict
from fnmatch import fnmatch
from typing import Optional

import h5py

from tomwer.core.process.control.datawatcher import status as datawatcherstatus
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.threads import LoopThread

from .datawatcherprocess import _DataWatchEmpty
from .edfdwprocess import (
    _DataWatcherProcessParseInfo,
    _DataWatcherProcessUserFilePattern,
    _DataWatcherProcessXML,
)
from .hdf5dwprocess import _BlissScanWatcherProcess, _DataWatcherProcessHDF5

logger = logging.getLogger(__name__)


class _DataWatcherStaticObserverMixIn:
    def __init__(
        self,
        scanID,
        obsMethod,
        srcPattern,
        destPattern,
        patternObs,
        observationRegistry,
    ):
        """

        :param str scanID: path to the scan
        :param obsMethod: method fo rdetecting end of acquisition
        :param Union[str,None] srcPattern:
        :param Union[str,None] destPattern:
        :param Union[str,None] patternObs:
        :param observationRegistry: what registry the observation should update
        """
        self.obsMethod = obsMethod
        self.srcPattern = srcPattern
        self.destPattern = destPattern
        self.patternObs = patternObs
        """The pattern to use for the DET_END_USER_ENTRY method"""
        self.setDirToObserve(scanID)
        self.dataWatcherProcess = self._getDataWatcherProcess()
        self.status = "not processing"
        self.observationRegistry = observationRegistry
        self.file_name_pattern = None

    def setProperties(self, properties):
        assert isinstance(properties, dict)
        if "scan" in properties:
            self.setDirToObserve(properties["scan"])

    def setDirToObserve(self, directory):
        self.path = directory
        if hasattr(self, "dataWatcherProcess") and self.dataWatcherProcess:
            self.dataWatcherProcess.RootDir = directory
            self.dataWatcherProcess.parsing_dir = ""

    def quit(self):
        if self.dataWatcherProcess is not None:
            self.dataWatcherProcess.quitting = True

    def run(self):
        look_for_hdf5_file = self.obsMethod in (
            datawatcherstatus.BLISS_SCAN_END,
            datawatcherstatus.NXtomo_END,
        )

        if not look_for_hdf5_file and not os.path.isdir(self.path):
            logger.info("can't observe %s, not a directory" % self.path)
            self.status = "failure"
            if self.observationRegistry:
                self.observationRegistry._updateStatus(
                    status=datawatcherstatus.OBSERVATION_STATUS[self.status],
                    scan=self.path,
                )
            self.validation = -1
            return

        if self.dataWatcherProcess.is_abort():
            if self.status != "aborted":
                logger.info("Acquisition %s has been aborted" % self.path)
                self.dataWatcherProcess._removeAcquisition(
                    scanID=self.path, reason="acquisition aborted by the user"
                )

                self.status = "aborted"
            if self.observationRegistry:
                self.observationRegistry._updateStatus(
                    status=datawatcherstatus.OBSERVATION_STATUS[self.status],
                    scan=self.path,
                )
            self.validation = -2
            return

        dataComplete = self.dataWatcherProcess.is_data_complete()

        if dataComplete is True:
            self.status = "acquisition ended"
            if self.observationRegistry:
                self.observationRegistry._updateStatus(
                    status=datawatcherstatus.OBSERVATION_STATUS[self.status],
                    scan=self.path,
                )
            self.validation = 1
        else:
            self.status = "waiting for acquisition ending"
            if self.observationRegistry:
                self.observationRegistry._updateStatus(
                    status=datawatcherstatus.OBSERVATION_STATUS[self.status],
                    scan=self.path,
                )
            self.validation = 0
        return

    def _getDataWatcherProcess(self):
        if self.obsMethod == datawatcherstatus.DET_END_XML:
            return _DataWatcherProcessXML(
                dataDir=self.path,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )
        elif self.obsMethod == datawatcherstatus.PARSE_INFO_FILE:
            return _DataWatcherProcessParseInfo(
                dataDir=self.path,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )
        elif self.obsMethod == datawatcherstatus.DET_END_USER_ENTRY:
            return _DataWatcherProcessUserFilePattern(
                dataDir=self.path,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
                pattern=self.patternObs,
            )
        elif self.obsMethod == datawatcherstatus.BLISS_SCAN_END:
            return _BlissScanWatcherProcess(
                dataDir=self.path,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )
        elif self.obsMethod == datawatcherstatus.NXtomo_END:
            return _DataWatcherProcessHDF5(
                dataDir=self.path,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )
        else:
            raise ValueError("requested observation method not recognized")

    def setSourceToDestinationPatterns(self, srcPattern, destPattern):
        """set the patterns to replace strings sequence in directories path.

        For example during acquisition in md05 acquisition files are stored
        in /lbsram/data/visitor/x but some information (as .info) files are
        stored in /data/visitor/x.
        So we would like to check information in both directories.
        Furthermore we would like that all file not in /data/visitor/x will be
        copied as soon as possible into /data/visitor/x (using RSyncManager)

        To do so we can define a srcPattern ('/lbsram' in our example) and
        destPattern : a string replacing to srcPattern in order to get both
        repositories. ('' in out example)
        If srcPattern or destPattern are setted to None then we won't apply
        this 'two directories' synchronization and check

        :param str srcPattern: the pattern to change by destPattern.
        :param str destPattern: the pattern that will replace srcPattern in the
            scan path
        """
        self.srcPattern = srcPattern
        self.destPattern = destPattern

    def setUnixFileNamePattern(self, pattern):
        self.file_name_pattern = pattern

    def setObservationMethod(self, obsMeth, info=None):
        """
        Set if we are looking for the .xml file

        :param str or tuple obsMeth:
        :param dict info: some extra information needed for some observation
                          method
        """
        assert info is None or type(info) is dict
        assert type(obsMeth) in (tuple, str)
        if type(obsMeth) is str:
            self.obsMethod = obsMeth
            if self.obsMethod == datawatcherstatus.DET_END_USER_ENTRY:
                assert "pattern" in info
                self.patternObs = info["pattern"]
            else:
                self.patternObs = None
        else:
            assert len(obsMeth) > 0
            assert type(obsMeth[0]) is str
            self.obsMethod = obsMeth[0]
            if self.obsMethod == datawatcherstatus.DET_END_USER_ENTRY:
                assert len(obsMeth) == 2
                assert type(obsMeth[1]) is dict
                assert "pattern" in obsMeth[1]
                self.patternObs = obsMeth[1]["pattern"]
            else:
                self.patternObs = None


class _DataWatcherStaticObserver(_DataWatcherStaticObserverMixIn, threading.Thread):
    """Observe one specific directory and signal when the state of this
    directory change"""

    def __init__(
        self,
        scanID,
        obsMethod,
        srcPattern,
        destPattern,
        patternObs,
        observationRegistry,
    ):
        threading.Thread.__init__(self)
        _DataWatcherStaticObserverMixIn.__init__(
            self,
            scanID=scanID,
            obsMethod=obsMethod,
            srcPattern=srcPattern,
            destPattern=destPattern,
            patternObs=patternObs,
            observationRegistry=observationRegistry,
        )


class _DataWatcherObserver_MixIn:
    """Thread launching the data watcher process (observation of acquisition)

    :param str headDir: the root dir to make to fin dacquisition
    :param int time_between_loops: seconds between observation threads
    :param bool startByOldest: if True then we parse folder from the oldest to
        the newest
    :param funcAdvancementHandler: handlers of the signals sended by data
                                   watcher (one for sigNbDirExplored and one
                                   for sigAdvanceExploration)
    :param str obsMethod: is True then will the creation of the xml file will
        notice the end of the acquisition. Otherwise we will look for .info
        file and for all .edf file to be copied
    :param str or None srcPattern: see data watcher
    :param str or None destPattern: see data watcher
    """

    latestScanReady = []

    def __init__(
        self,
        obsMethod,
        time_between_loops,
        observationClass,
        headDir=None,
        startByOldest=False,
        srcPattern=None,
        destPattern=None,
        ignoredFolders=None,
        file_name_pattern=None,
    ):
        self.observations = observationClass()
        self.observations.ignoredFolders = (
            [] if ignoredFolders is None else ignoredFolders
        )
        """dict of observer on one specific scan. Key is the directory,
        value the :class:`DataWatcherFixObserver`"""
        self.setHeadFolder(headDir)
        self.setObservationMethod(obsMethod)
        self.srcPattern = srcPattern
        self.destPattern = destPattern
        self._file_name_pattern = file_name_pattern
        self.dataWatcherProcess = None
        self._patternObs = None
        """The pattern to use for the DET_END_USER_ENTRY method"""
        self.lock = threading.Lock()
        self.restartEvent = None
        self._firstRun = True

    @property
    def file_name_pattern(self) -> Optional[str]:
        return self._file_name_pattern

    @file_name_pattern.setter
    def file_name_pattern(self, pattern: Optional[str]):
        if pattern is None or pattern.lstrip(" ").rstrip(" ") == "":
            self._file_name_pattern = None
        else:
            self._file_name_pattern = pattern.lstrip(" ").rstrip(" ")

    def resetObservations(self):
        self.observations.reset()

    def registerRestartProcess(self, event):
        self.restartEvent = event

    def run(self):
        look_for_hdf5_file = self.obsMethod in (
            datawatcherstatus.BLISS_SCAN_END,
            datawatcherstatus.NXtomo_END,
        )

        def process_observations(file_path):
            do_observation = (
                self.observations.isObserving(file_path) is False
                and (
                    (look_for_hdf5_file and h5py.is_hdf5(file_path))
                    or (self.dataWatcherProcess._isScanDirectory(file_path))
                )
                and file_path not in self.observations.ignoredFolders
            )
            if (
                do_observation
                and self.file_name_pattern is not None
                and not fnmatch(os.path.basename(file_path), self.file_name_pattern)
            ):
                do_observation = False
            if do_observation:
                # do the observation
                self.observe(file_path)

            if os.path.isdir(file_path):
                try:
                    for f in os.listdir(file_path):
                        full_file_path = os.path.join(file_path, f)
                        if os.path.isdir(full_file_path) or (
                            look_for_hdf5_file and h5py.is_hdf5(full_file_path)
                        ):
                            # parse files to check if some observations should be done
                            process_observations(full_file_path)
                except Exception:
                    pass

        if self._firstRun or (
            self.restartEvent is not None and self.restartEvent.isSet()
        ):
            self._firstRun = False
            if self.restartEvent:
                self.restartEvent.clear()

            if not os.path.isdir(self.headDir):
                logger.warning("can't observe %s, not a directory" % self.headDir)
                return
            self._check_scans_ready()  # pylint: disable=E1101
            self.dataWatcherProcess = self._getDataWatcherProcess()
            process_observations(self.headDir)
            self._processObservation()
            self._check_scans_ready()  # pylint: disable=E1101
        else:
            time.sleep(0.5)

    def _processObservation(self):
        threads = list(self.observations.dict.values())
        for thread in threads:
            thread.start()

    def quit(self):
        threads = list(self.observations.dict.values())
        for thread in threads:
            thread.quitting = True

        if self.dataWatcherProcess is not None:
            self.dataWatcherProcess.quitting = True
            self.lastObsDir = None

    def setUnixFileNamePattern(self, pattern):
        self.file_name_pattern = pattern

    def setObservationMethod(self, obsMeth, info=None):
        """
        Set if we are looking for the .xml file from .edf observation or any
        .hdf5 file

        :param str or tuple obsMeth:
        :param dict info: some extra information needed for some observation
                          method
        """
        assert info is None or type(info) is dict
        assert type(obsMeth) in (tuple, str)
        if type(obsMeth) is str:
            self.obsMethod = obsMeth
            if self.obsMethod == datawatcherstatus.DET_END_USER_ENTRY:
                assert "pattern" in info
                self._patternObs = info["pattern"]
            else:
                self._patternObs = None
        else:
            assert len(obsMeth) > 0
            assert type(obsMeth[0]) is str
            self.obsMethod = obsMeth[0]
            if self.obsMethod == datawatcherstatus.DET_END_USER_ENTRY:
                assert len(obsMeth) == 2
                assert type(obsMeth[1]) is dict
                assert "pattern" in obsMeth[1]
                self._patternObs = obsMeth[1]["pattern"]
            else:
                self._patternObs = None

        for _, thread in self.observations.dict.items():
            thread.setObservationMethod(obsMeth, info=info)

    def setHeadFolder(self, headDir):
        assert type(headDir) in (type(None), str)
        self.headDir = headDir

    def setSourceToDestinationPatterns(self, srcPattern, destPattern):
        """set the patterns to replace strings sequence in directories path.

        For example during acquisition in md05 acquisition files are stored
        in /lbsram/data/visitor/x but some information (as .info) files are
        stored in /data/visitor/x.
        So we would like to check information in both directories.
        Furthermore we would like that all file not in /data/visitor/x will be
        copied as soon as possible into /data/visitor/x (using RSyncManager)

        To do so we can define a srcPattern ('/lbsram' in our example) and
        destPattern : a string replacing to srcPattern in order to get both
        repositories. ('' in out example)
        If srcPattern or destPattern are setted to None then we won't apply
        this 'two directories' synchronization and check

        :param str srcPattern: the pattern to change by destPattern.
        :param str destPattern: the pattern that will replace srcPattern in the
            scan path
        """
        self.srcPattern = srcPattern
        self.destPattern = destPattern
        for _, thread in self.observations.items():
            thread.setSourceToDestinationPatterns(self.srcPattern, self.destPattern)

    def observe(self, scanID):
        self.observations.add(self._getObserver(scanID))  # pylint: disable=E1101

    def cancelObservation(self, scanID):
        if self.observations.isObserving(scanID) is False:
            logger.warning(
                "Can't cancel observation on %s, no observation " "registred" % scanID
            )
            return

        self.observations.ignoredFolders.append(scanID)
        if scanID in self.observations.dict:
            self.observations.remove(self.observations.dict[scanID])

    def isObserve(self, scanID):
        return self.observations.isObserving(scanID)

    def _getDataWatcherProcess(self):
        if self.obsMethod == datawatcherstatus.NXtomo_END:
            return _DataWatcherProcessHDF5(
                dataDir=self.headDir,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )
        elif self.obsMethod == datawatcherstatus.BLISS_SCAN_END:
            return _BlissScanWatcherProcess(
                dataDir=self.headDir,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )
        elif EDFTomoScan.directory_contains_scan(
            self.headDir, src_pattern=self.srcPattern, dest_pattern=self.destPattern
        ):
            if self.obsMethod == datawatcherstatus.DET_END_XML:
                return _DataWatcherProcessXML(
                    dataDir=self.headDir,
                    srcPattern=self.srcPattern,
                    destPattern=self.destPattern,
                )
            elif self.obsMethod == datawatcherstatus.PARSE_INFO_FILE:
                return _DataWatcherProcessParseInfo(
                    dataDir=self.headDir,
                    srcPattern=self.srcPattern,
                    destPattern=self.destPattern,
                )
            elif self.obsMethod == datawatcherstatus.DET_END_USER_ENTRY:
                return _DataWatcherProcessUserFilePattern(
                    dataDir=self.headDir,
                    srcPattern=self.srcPattern,
                    destPattern=self.destPattern,
                    pattern=self._patternObs,
                )
        else:
            return _DataWatchEmpty(
                dataDir=self.headDir,
                srcPattern=self.srcPattern,
                destPattern=self.destPattern,
            )


class _DataWatcherObserver(_DataWatcherObserver_MixIn, LoopThread):
    """DataWatcherObserver using Threads"""

    scanReadyEvent = threading.Event()
    """Event set when at least one new scan is found"""

    def __init__(
        self,
        obsMethod,
        time_between_loops,
        observationClass,
        headDir=None,
        startByOldest=False,
        srcPattern=None,
        destPattern=None,
        ignoredFolders=None,
    ):
        LoopThread.__init__(self, time_between_loops=time_between_loops)
        _DataWatcherObserver_MixIn.__init__(
            self,
            obsMethod=obsMethod,
            observationClass=observationClass,
            headDir=headDir,
            startByOldest=startByOldest,
            srcPattern=srcPattern,
            destPattern=destPattern,
            ignoredFolders=ignoredFolders,
            time_between_loops=time_between_loops,
        )

    def _check_scans_ready(self):
        with self.observations.lock:
            if self.observations.scanReadyEvent.isSet():
                for scanReady in self.observations._latestScanReady:
                    self._signalScanReady(scanID=scanReady)
                self.observations.scanReadyEvent.clear()
                self.observations._latestScanReady.clear()

    def _signalScanReady(self, scanID):
        with self.lock:
            self.scanReadyEvent.set()
            self.latestScanReady.append(scanID)

    def _getObserver(self, scanID):
        return _DataWatcherStaticObserver(
            scanID=scanID,
            obsMethod=self.obsMethod,
            srcPattern=self.srcPattern,
            destPattern=self.destPattern,
            patternObs=self._patternObs,
            observationRegistry=self.observations,
        )

    def waitForObservationFinished(self, timeOut=10):
        threads = list(self.observations.dict.values())
        for thread in threads:
            if thread.is_alive():
                thread.join(timeOut)


class _OngoingObservation(object):
    """
    Simple container of observed directory
    """

    scanReadyEvent = threading.Event()
    """Event set a finished acquisition is detected"""
    obsAddedEvent = threading.Event()
    """Event set when an observation is added"""
    obsRemovedEvent = threading.Event()
    """Event set when an observation is removed"""
    obsStatusReceivedEvent = threading.Event()
    """Event set when receiving a new observation status"""

    def __init__(self):
        self.dict = OrderedDict()
        """keys are path, value are observers"""
        self.ignoredFolders = []
        self.lock = threading.Lock()
        self._latestObsEnded = []
        """list of latest observation with scan"""
        self._latestObsAdded = []
        """list of latest removed observations with scan"""
        self._latestObsStatusChanged = []
        """list of latest modify observation with (scan, status)"""
        self._latestScanReady = []
        """list of latest scan ready"""

    def add(self, observer):
        if self.isObserving(observer.path) is False:
            with self.lock:
                assert isinstance(observer, _DataWatcherStaticObserverMixIn)
                self.dict[observer.path] = observer
                self.obsAddedEvent.set()
                self._latestObsEnded.append(observer.path)

    def remove(self, observer):
        assert isinstance(observer, _DataWatcherStaticObserverMixIn)
        if self.isObserving(observer.path) is True:
            # observer.sigStatusChanged.disconnect(self._updateStatus)
            observer.quit()
            del self.dict[observer.path]
            with self.lock:
                self.obsRemovedEvent.set()
                self._latestObsEnded.append(observer.path)

    def _updateStatus(self, status, scan):
        if self.isObserving(scan) is True:
            with self.lock:
                self.obsStatusReceivedEvent.set()
                self._latestObsStatusChanged.append(
                    (scan, datawatcherstatus.DICT_OBS_STATUS[status])
                )
            if status == datawatcherstatus.OBSERVATION_STATUS["acquisition ended"]:
                self._acquisition_ended(scan)
            if status in (
                datawatcherstatus.OBSERVATION_STATUS["failure"],
                datawatcherstatus.OBSERVATION_STATUS["aborted"],
            ):
                observer = self.dict[scan]
                self.remove(observer)

    def _acquisition_ended(self, scanID):
        with self.lock:
            self.ignoredFolders.append(scanID)
            observer = self.dict[scanID]
        self.remove(observer)
        # TODO : disconnect the thread and delete it if finisehd
        #  add the scan to the one to ignore. Those ignore
        # should be removed at each start and stop of the observation
        with self.lock:
            self.scanReadyEvent.set()
            self._latestScanReady.append(scanID)

    def isObserving(self, scan):
        return scan in self.dict

    def reset(self):
        for scanID, observer in self.dict:
            # observer.sigStatusChanged.disconnect(self._updateStatus)
            observer.quit()
        self.dict = {}

    def __len__(self):
        return len(self.dict)

    def __str__(self):
        return str(self.dict.keys())



import os
from silx.gui import qt
from silx.utils.enum import Enum as _Enum
from tomwer.core.process.icat.gallery import OutputFormat
from tomwer.gui.qlefilesystem import QLFileSystem
from tomwer.io.utils import get_default_directory
from tomwer.core.utils.dictutils import concatenate_dict
from tomwer.core.process.icat.gallery import Binning

from .publish import PublishProcessedDataWidget


class _GalleryOutputDir(qt.QGroupBox):
    class OutputDirMode(_Enum):
        DATASET_GALLERY = "dataset gallery"
        PROPOSAL_GALLERY = "proposal GALLERY"
        CUSTOM = "custom"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())
        # dataset gallery option
        self._datasetGalleryQRB = qt.QRadioButton(
            self.OutputDirMode.DATASET_GALLERY.value, self
        )
        self.layout().addWidget(self._datasetGalleryQRB, 0, 0, 1, 4)
        # proposal gallery option
        self._proposalGalleryQRB = qt.QRadioButton(
            self.OutputDirMode.PROPOSAL_GALLERY.value, self
        )
        self.layout().addWidget(self._proposalGalleryQRB, 1, 0, 1, 4)

        # other option
        self._otherQRB = qt.QRadioButton(self.OutputDirMode.CUSTOM.value, self)
        self._otherQRB.setCheckable(True)
        self.layout().addWidget(self._otherQRB, 2, 0, 1, 4)

        self._otherQLE = QLFileSystem(
            "", self, filters=qt.QDir.NoDotAndDotDot | qt.QDir.Dirs
        )
        self.layout().addWidget(self._otherQLE, 3, 1, 1, 2)
        style = qt.QApplication.style()
        icon_opendir = style.standardIcon(qt.QStyle.SP_DirOpenIcon)
        self._selectOtherQLE = qt.QPushButton(icon_opendir, "", self)
        self._selectOtherQLE.setIcon(icon_opendir)
        self.layout().addWidget(self._selectOtherQLE, 3, 3, 1, 1)

        # button group
        self._buttonGroup = qt.QButtonGroup()
        self._buttonGroup.setExclusive(True)
        self._buttonGroup.addButton(self._datasetGalleryQRB)
        self._buttonGroup.addButton(self._proposalGalleryQRB)
        self._buttonGroup.addButton(self._otherQRB)

        # vertical spacer
        self._vSpacer = qt.QWidget(self)
        self._vSpacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._vSpacer, 99, 0, 1, 1)

        # set up GUI
        self._datasetGalleryQRB.setChecked(True)
        self._modeChanged()

        # connect signal / slot
        self._buttonGroup.buttonClicked.connect(self._modeChanged)
        self._selectOtherQLE.released.connect(self._selectOutput)

    def _selectOutput(self):  # pragma: no cover
        defaultDirectory = self._selectOtherQLE.text()
        if not os.path.isdir(defaultDirectory):
            defaultDirectory = get_default_directory()

        dialog = qt.QFileDialog(self, directory=defaultDirectory)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return

        self._selectOtherQLE.setText(dialog.selectedFiles()[0])

    def getOutputFolderMode(self):
        if self._datasetGalleryQRB.isChecked():
            return _GalleryOutputDir.OutputDirMode.DATASET_GALLERY
        elif self._proposalGalleryQRB.isChecked():
            return _GalleryOutputDir.OutputDirMode.PROPOSAL_GALLERY
        elif self._otherQRB.isChecked():
            return _GalleryOutputDir.OutputDirMode.CUSTOM
        else:
            raise NotImplementedError

    def setOutputFolderMode(self, mode):
        mode = _GalleryOutputDir.OutputDirMode.from_value(mode)
        if mode is _GalleryOutputDir.OutputDirMode.PROPOSAL_GALLERY:
            self._proposalGalleryQRB.setChecked(True)
        elif mode is _GalleryOutputDir.OutputDirMode.DATASET_GALLERY:
            self._datasetGalleryQRB.setChecked(True)
        elif mode is _GalleryOutputDir.OutputDirMode.CUSTOM:
            self._otherQRB.setChecked(True)
        else:
            raise NotImplementedError
        self._modeChanged()

    def getOtherOutputDir(self):
        return self._otherQLE.text()

    def setOtherOutputDir(self, path: str):
        self._otherQLE.setText(path)

    def _modeChanged(self, *args, **kwargs):
        custom_active = self._otherQRB.isChecked()
        self._otherQLE.setEnabled(custom_active)
        self._selectOtherQLE.setEnabled(custom_active)


class GalleryWidget(qt.QWidget):
    """Widget to let the user define the output location of the screenshots"""

    sigConfigChanged = qt.Signal()
    """emit when the configuration has changed"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())

        # screenshot precision
        self._precisonQCB = qt.QComboBox(self)
        self._precisonQCB.addItem("uint8")
        self.layout().addRow("precision", self._precisonQCB)
        # binning
        self._binningQCB = qt.QComboBox(self)
        self._binningQCB.addItems(Binning.values())
        self.layout().addRow("binning", self._binningQCB)
        self._binningQCB.setCurrentText(Binning.SIXTEEN_BY_SIXTEEN.value)
        self._binningQCB.setToolTip(
            "To speed up display of the gallery at the data portal side it is highly recommended to bin screenshots"
        )  # recommanded size: 5ko for the entire gallery
        # output format
        self._outputFormat = qt.QComboBox(self)
        self._outputFormat.addItems(OutputFormat.values())
        self.layout().addRow("outoutput_location_modeput format", self._outputFormat)
        # gallery output dir
        self._outputDirWidget = _GalleryOutputDir(self)
        self._outputLocationLabel = qt.QLabel("output location", self)
        self.layout().addRow(self._outputLocationLabel, self._outputDirWidget)
        self._outputDirWidget.hide()
        self._outputLocationLabel.hide()
        # overwrite
        self._overwriteCB = qt.QCheckBox("overwrite", self)
        self._overwriteCB.setChecked(True)
        self.layout().addRow(self._overwriteCB)
        # publishing setting
        self._publishConfig = PublishProcessedDataWidget(self)
        self._publisherGB = qt.QGroupBox("icat info")
        self._publisherGB.setLayout(qt.QVBoxLayout())
        self._publisherGB.layout().addWidget(self._publishConfig)
        self.layout().addRow(self._publisherGB)
        # connect signal / slot
        self._outputFormat.currentIndexChanged.connect(self._configChanged)
        self._overwriteCB.toggled.connect(self._configChanged)
        self._publishConfig.sigConfigChanged.connect(self.sigConfigChanged)
        self._binningQCB.currentIndexChanged.connect(self._configChanged)

    def getOutputFormat(self) -> OutputFormat:
        return OutputFormat.from_value(self._outputFormat.currentText())

    def setOutputFormat(self, format: OutputFormat):
        format = OutputFormat.from_value(format)
        self._outputFormat.setCurrentText(format.value)

    def getBinning(self) -> Binning:
        return Binning.from_value(self._binningQCB.currentText())

    def setBinning(self, binning: Binning):
        binning = Binning.from_value(binning)
        self._binningQCB.setCurrentText(binning.value)

    def getConfiguration(self):
        return concatenate_dict(
            self._publishConfig.getConfiguration(),
            {
                "output_format": self.getOutputFormat().value,
                "output_location_mode": self._outputDirWidget.getOutputFolderMode().value,
                "custom_output": self._outputDirWidget.getOtherOutputDir(),
                "overwrite": self._overwriteCB.isChecked(),
                "binning": self.getBinning().value,
            },
        )

    def setConfiguration(self, config: dict):
        self._publishConfig.setConfiguration(config)
        output_format = config.get("output_format", None)
        if output_format is not None:
            self.setOutputFormat(output_format)

        output_location_mode = config.get("output_location_mode", None)
        if output_location_mode is not None:
            self._outputDirWidget.setOutputFolderMode(output_location_mode)

        custom_output = config.get("custom_output", None)
        if custom_output is not None:
            self._outputDirWidget.setOtherOutputDir(custom_output)

        overwrite = config.get("overwrite", None)
        if overwrite is not None:
            overwrite = overwrite in (True, "True", 1)
            self._overwriteCB.setChecked(overwrite)

        binning = config.get("binning", None)
        if binning is not None:
            self.setBinning(binning=binning)

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/02/2020"


import logging
import os
import typing

from silx.gui import qt

from tomwer.core.process.reconstruction.output import NabuOutputFileFormat
from tomwer.core.process.reconstruction.nabu.utils import _NabuStages
from tomwer.core.process.reconstruction.output import ProcessDataOutputDirMode
from tomwer.gui.qlefilesystem import QLFileSystem
from tomwer.gui.reconstruction.nabu.nabuconfig.base import _NabuStageConfigBase
from tomwer.io.utils import get_default_directory

try:
    import glymur  # noqa #F401 needed for later possible lazy loading
except ImportError:
    has_glymur = False
else:
    has_glymur = True

_logger = logging.getLogger(__name__)


class QNabuFileFormatComboBox(qt.QComboBox):
    def __init__(
        self, parent: typing.Optional[qt.QWidget] = ..., filter_formats=tuple()
    ) -> None:
        """
        :param filter_format: if provided the given file format won't have an item on the ComboBox
        """
        super().__init__(parent=parent)
        if not isinstance(filter_formats, tuple):
            raise TypeError(
                f"filter_format should be a tuple. Get {type(filter_formats)} instead"
            )
        filter_formats = [
            NabuOutputFileFormat.from_value(file_format)
            for file_format in filter_formats
        ]

        for ff in NabuOutputFileFormat:
            if ff in filter_formats:
                continue
            if ff is NabuOutputFileFormat.JP2K:
                if not has_glymur:
                    _logger.warning(
                        "could not load jp2k format, glymur and OpenJPEG requested"
                    )
                else:
                    from glymur import version

                    if version.openjpeg_version < "2.3.0":
                        _logger.warning(
                            "You must have at least version 2.3.0 of OpenJPEG "
                            "in order to write jp2k images."
                        )
                    else:
                        self.addItem(ff.value)
            else:
                self.addItem(ff.value)


class QNabuFileFormatComboBoxIgnoreWheel(QNabuFileFormatComboBox):
    def wheelEvent(self, e: qt.QWheelEvent) -> None:
        pass


class NabuOutputLocationWidget(qt.QGroupBox):
    sigOutputChanged = qt.Signal()
    """Emit when location changed"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(title="output folder", *args, **kwargs)
        self.setLayout(qt.QGridLayout())

        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        # in scan folder
        self._inScanFolderRB = qt.QRadioButton(
            ProcessDataOutputDirMode.IN_SCAN_FOLDER.value, self
        )
        self._inScanFolderRB.setToolTip(
            "Reconstruction will be saved at the same level as the acquisition folder. (near the NXtomo file (.nx) or under the spec acquisition folder)"
        )
        self.layout().addWidget(self._inScanFolderRB, 1, 0, 1, 1)
        # in processed data dir
        self._processedDataDirRB = qt.QRadioButton(
            ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER.value, self
        )
        self._inScanFolderRB.setToolTip(
            "Reconstruction will be saved under the PROCESSED_DATA/dataset folder. if exists else under the scan folder."
        )
        self.layout().addWidget(self._processedDataDirRB, 2, 0, 1, 1)
        # raw data dir
        self._rawDataDirRB = qt.QRadioButton(
            ProcessDataOutputDirMode.RAW_DATA_FOLDER.value, self
        )
        self._rawDataDirRB.setToolTip(
            "Reconstruction will be saved under the RAW_DATA/dataset folder. if exists else under the scan folder."
        )
        self.layout().addWidget(self._rawDataDirRB, 3, 0, 1, 1)
        # other dir
        self._otherDirRB = qt.QRadioButton(ProcessDataOutputDirMode.OTHER.value, self)
        self._otherDirRB.setToolTip(
            "Reconstruction will be saved under user provided folder."
        )
        self.layout().addWidget(self._otherDirRB, 4, 0, 1, 1)
        self._outputDirQLE = QLFileSystem(
            "", self, filters=qt.QDir.NoDotAndDotDot | qt.QDir.Dirs
        )
        self.layout().addWidget(self._outputDirQLE, 4, 1, 1, 1)
        style = qt.QApplication.style()
        icon_opendir = style.standardIcon(qt.QStyle.SP_DirOpenIcon)
        self._selectOutputPB = qt.QPushButton(icon_opendir, "", self)
        self._selectOutputPB.setIcon(icon_opendir)
        general_tooltip = (
            "You can enter a string with some keywords like {my_keyword}. Those will be interpreted during runtime according to scan metadata. Possible keywords are:"
            + "\n - 'scan_dir_name': returns name of the directory containing the acquisition (! not a path !)"
            + "\n - 'scan_basename': returns basename of the directory containing the acquisition"
            + "\n - 'scan_parent_dir_basename': returns basename of the PARENT directory containing the acquisition"
        )

        self._selectOutputPB.setToolTip(general_tooltip)
        self._outputDirQLE.setToolTip(general_tooltip)
        self.layout().addWidget(self._selectOutputPB, 4, 2, 1, 1)

        # set up:
        self._inScanFolderRB.setChecked(True)

        # connect signal / slot
        self._inScanFolderRB.toggled.connect(self._outputModeChanged)
        self._processedDataDirRB.toggled.connect(self._outputModeChanged)
        self._rawDataDirRB.toggled.connect(self._outputModeChanged)
        self._otherDirRB.toggled.connect(self._outputModeChanged)
        self._selectOutputPB.released.connect(self._selectOutput)
        self._inScanFolderRB.setChecked(True)
        self._outputModeChanged()

    def getOutputDirMode(self) -> ProcessDataOutputDirMode:
        if self._inScanFolderRB.isChecked():
            return ProcessDataOutputDirMode.IN_SCAN_FOLDER
        elif self._processedDataDirRB.isChecked():
            return ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER
        elif self._rawDataDirRB.isChecked():
            return ProcessDataOutputDirMode.RAW_DATA_FOLDER
        elif self._otherDirRB.isChecked():
            return ProcessDataOutputDirMode.OTHER

    def setOutputDirMode(
        self, mode: typing.Union[ProcessDataOutputDirMode, str]
    ) -> None:
        mode = ProcessDataOutputDirMode.from_value(mode)
        if mode is ProcessDataOutputDirMode.IN_SCAN_FOLDER:
            self._inScanFolderRB.setChecked(True)
        elif mode is ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER:
            self._processedDataDirRB.setChecked(True)
        elif mode is ProcessDataOutputDirMode.RAW_DATA_FOLDER:
            self._rawDataDirRB.setChecked(True)
        elif mode is ProcessDataOutputDirMode.OTHER:
            self._otherDirRB.setChecked(True)

    def _outputModeChanged(self, *args, **kwargs):
        outputMode = self.getOutputDirMode()
        self._outputDirQLE.setVisible(outputMode is ProcessDataOutputDirMode.OTHER)
        self._selectOutputPB.setVisible(outputMode is ProcessDataOutputDirMode.OTHER)
        self.sigOutputChanged.emit()

    def _selectOutput(self):  # pragma: no cover
        defaultDirectory = self._outputDirQLE.text()
        if not os.path.isdir(defaultDirectory):
            defaultDirectory = get_default_directory()

        dialog = qt.QFileDialog(self, directory=defaultDirectory)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return

        self._outputDirQLE.setText(dialog.selectedFiles()[0])

    def getOutputDir(self):
        """

        :return: None if the default output directory is selected else
                 return path to the directory
        """
        if self._otherDirRB.isChecked():
            return self._outputDirQLE.text()
        else:
            return None

    def setOutputDir(self, output_dir):
        if output_dir in (None, ""):
            pass
        else:
            self._outputDirQLE.setText(output_dir)


class _NabuOutputConfig(_NabuStageConfigBase, qt.QWidget):
    """
    Widget to define the output configuration of nabu
    """

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        _NabuStageConfigBase.__init__(self, stage=_NabuStages.POST)
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        # output dir
        self._output_dir_widget = NabuOutputLocationWidget(parent=self)
        self.layout().addWidget(self._output_dir_widget, 0, 0, 1, 4)
        self.registerWidget(self._output_dir_widget, "advanced")

        # file format
        self._outputFileFormatLabel = qt.QLabel("output file format:", self)
        self.layout().addWidget(self._outputFileFormatLabel, 1, 0, 1, 1)
        self._fileFormatCB = QNabuFileFormatComboBoxIgnoreWheel(self)
        self.layout().addWidget(self._fileFormatCB, 1, 2, 1, 2)
        self.registerWidget(self._outputFileFormatLabel, "optional")
        self.registerWidget(self._fileFormatCB, "optional")

        # file per group
        self._filePerGroupLabel = qt.QLabel("frame per group:", self)
        self.layout().addWidget(self._filePerGroupLabel, 2, 0, 1, 1)
        self._framePerGroup = qt.QSpinBox(self)
        self._framePerGroup.setMinimum(100)
        self._framePerGroup.setSingleStep(50)
        self._framePerGroup.setMaximum(10000)
        # not managed for now so hide
        self._filePerGroupLabel.hide()
        self._framePerGroup.hide()
        self.layout().addWidget(self._framePerGroup, 2, 2, 1, 2)

        # spacer for style
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 200, 0, 1, 1)

        # set up
        self.setFramePerGroup(100)
        self.setFileformat("hdf5")

        # connect signal / slot
        self._output_dir_widget._outputDirQLE.editingFinished.connect(
            self._outputDirChanged
        )
        self._output_dir_widget.sigOutputChanged.connect(self._outputDirChanged)
        self._fileFormatCB.currentTextChanged.connect(self._fileFormatChanged)
        self._framePerGroup.valueChanged.connect(self._framePerGroupChanged)

    def _outputDirChanged(self):
        self.sigConfChanged.emit("location")

    def _fileFormatChanged(self):
        self.sigConfChanged.emit("file_format")

    def _framePerGroupChanged(self):
        self.sigConfChanged.emit("frames_per_group")

    def getOutputDir(self):
        return self._output_dir_widget.getOutputDir()

    def setOutputDir(self, dir):
        return self._output_dir_widget.setOutputDir(dir)

    def getOutputdirMode(self):
        return self._output_dir_widget.getOutputDirMode()

    def setOutputdirMode(self, mode):
        return self._output_dir_widget.setOutputDirMode(mode=mode)

    def getFileFormat(self) -> NabuOutputFileFormat:
        return NabuOutputFileFormat.from_value(self._fileFormatCB.currentText())

    def setFileformat(self, file_format):
        file_format = NabuOutputFileFormat.from_value(file_format)
        index = self._fileFormatCB.findText(file_format.value)
        self._fileFormatCB.setCurrentIndex(index)

    def getFramePerGroup(self):
        return self._framePerGroup.value()

    def setFramePerGroup(self, n_frames):
        self._framePerGroup.setValue(n_frames)

    def getConfiguration(self):
        return {
            "file_format": self.getFileFormat().value,
            "location": self.getOutputDir() or "",
            "output_dir_mode": self.getOutputdirMode().value,
            # 'frames_per_group': self.getFramePerGroup(),
        }

    def setConfiguration(self, config):
        if "file_format" in config:
            self.setFileformat(config["file_format"])
        location = config.get("location", None)
        if location == "":
            location = None
        if location is not None:
            self.setOutputDir(location)
        if "frames_per_group" in config:
            self.setFramePerGroup(int(config["frames_per_group"]))

        # definition of default_output_dir_mode ensure backward compatibility
        default_output_dir_mode = (
            None if location is None else ProcessDataOutputDirMode.OTHER
        )
        output_dir_mode = config.get("output_dir_mode", default_output_dir_mode)
        if output_dir_mode is not None:
            self.setOutputdirMode(output_dir_mode)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""module defining dedicated completer"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/03/2022"


from typing import Optional

from silx.gui import qt


class _CodeCompleter(qt.QCompleter):
    ConcatenationRole = qt.Qt.UserRole + 1

    def __init__(self, data, parent=None):
        super().__init__(parent)
        self.create_model(data)

    def pathFromIndex(self, ix):
        return ix.data(_CodeCompleter.ConcatenationRole)

    def create_model(self, data):
        def addItems(parent, elements, t=""):
            for text in elements:
                item = qt.QStandardItem(text)
                data = t + "." + text if t else text
                item.setData(data, _CodeCompleter.ConcatenationRole)
                parent.appendRow(item)

        model = qt.QStandardItemModel(self)
        addItems(model, data)
        self.setModel(model)


class _LineEditWithCompleter(qt.QLineEdit):
    def __init__(self, completer, parent=None, data_type="url"):
        super().__init__(parent)
        self._completer = completer

    def keyPressEvent(self, key_event):
        # add Ctr + space for completion
        modifiers = key_event.modifiers()
        if qt.Qt.KeyboardModifier.ControlModifier == modifiers:
            if key_event.key() == qt.Qt.Key.Key_Space:
                self._completer.complete(self.rect())

        super().keyPressEvent(key_event)


class _TextCompleterWidget(qt.QWidget):
    """
    Simple widget to ease text research.
    """

    sigTextSelected = qt.Signal(str)

    def __init__(self, texts: tuple, parent=None) -> None:
        if not isinstance(texts, tuple):
            raise TypeError("texts is expected to be a tuple.")
        super().__init__(parent)
        layout = qt.QVBoxLayout()
        self.setLayout(layout)

        self._completer = _CodeCompleter(texts, self)

        self._lineEdit = _LineEditWithCompleter(parent=self, completer=self._completer)
        self._lineEdit.setCompleter(self._completer)
        layout.addWidget(self._lineEdit)

    def setText(self, text: str) -> None:
        self._lineEdit.setText(text)

    def text(self) -> str:
        return self._lineEdit.text()


class UrlCompleterDialog(qt.QDialog):
    def __init__(
        self, urls: tuple, parent=None, current_url: Optional[str] = None
    ) -> None:
        super().__init__(parent)
        self._urls = urls
        self.setWindowTitle("Url research")

        self.setLayout(qt.QVBoxLayout())
        self._completerWidget = _TextCompleterWidget(texts=urls, parent=self)
        self._completerWidget.setToolTip(
            "You can complete url. Use ctr + space to complete"
        )
        if current_url is not None:
            self._completerWidget.setText(current_url)
        self.layout().addWidget(self._completerWidget)

        types = qt.QDialogButtonBox.Cancel | qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Cancel).released.connect(self.reject)
        self._buttons.button(qt.QDialogButtonBox.Ok).released.connect(self.accept)
        self._completerWidget._lineEdit.textChanged.connect(
            self._updateOkButtonAvailability
        )

    def _updateOkButtonAvailability(self, *args, **kwargs):
        ok_enable = self._completerWidget.text() in self._urls
        self._buttons.button(qt.QDialogButtonBox.Ok).setEnabled(ok_enable)

    def selected_url(self):
        return self._completerWidget.text()

    def sizeHint(self):
        return qt.QSize(600, 100)



import platform
from silx.gui import qt
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.utils.buttons import PadlockButton


class PublishProcessedDataWidget(qt.QWidget):
    sigConfigChanged = qt.Signal()
    """emit when the configuration changed"""

    KNOW_BEAMLINES = sorted(
        ("bm05", "bm18", "id11", "id15a", "id16a", "id16b", "id17", "id19")
    )

    _TOOLTIP_PAD_LOCKS = "when receive a new scan the value will be updated if valid informations are found. If the field is lock then no update will be done automatically."

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())

        # beamline
        self._beamlineCB = qt.QComboBox(self)
        self._beamlineCB.setEditable(True)
        self._beamlineCB.addItems(self.KNOW_BEAMLINES)
        self._beamlineCB.lineEdit().setPlaceholderText("beamline name")
        self.layout().addWidget(
            qt.QLabel("beamline", self),
            0,
            0,
            1,
            1,
        )
        self.layout().addWidget(
            self._beamlineCB,
            0,
            1,
            1,
            1,
        )
        self._beamlinePLB = PadlockButton(self)
        self._beamlinePLB.setToolTip(self._TOOLTIP_PAD_LOCKS)
        self.layout().addWidget(
            self._beamlinePLB,
            0,
            2,
            1,
            1,
        )

        # proposal
        self._proposalQLE = qt.QLineEdit("", self)
        self._proposalQLE.setPlaceholderText("proposal name")
        self.layout().addWidget(
            qt.QLabel("proposal", self),
            1,
            0,
            1,
            1,
        )
        self.layout().addWidget(
            self._proposalQLE,
            1,
            1,
            1,
            1,
        )
        self._proposalPLB = PadlockButton(self)
        self._proposalPLB.setToolTip(self._TOOLTIP_PAD_LOCKS)
        self.layout().addWidget(
            self._proposalPLB,
            1,
            2,
            1,
            1,
        )

        # dataset
        self._datasetQLE = qt.QLineEdit("", self)
        self._datasetQLE.setPlaceholderText("dataset name")
        self.layout().addWidget(
            qt.QLabel("dataset", self),
            2,
            0,
            1,
            1,
        )
        self.layout().addWidget(
            self._datasetQLE,
            2,
            1,
            1,
            1,
        )
        self._datasetPLB = PadlockButton(self)
        self._datasetPLB.setToolTip(self._TOOLTIP_PAD_LOCKS)
        self.layout().addWidget(
            self._datasetPLB,
            2,
            2,
            1,
            1,
        )

        # set up
        default_beamline = self.getDefaultBeamline()
        if default_beamline is not None:
            self._beamlineCB.setCurrentText(default_beamline)

        # connect signal / slot
        self._beamlinePLB.toggled.connect(self._configChanged)
        self._proposalPLB.toggled.connect(self._configChanged)
        self._datasetPLB.toggled.connect(self._configChanged)
        self._beamlineCB.currentTextChanged.connect(self._configChanged)
        self._proposalQLE.textEdited.connect(self._configChanged)
        self._datasetQLE.textEdited.connect(self._configChanged)

    @staticmethod
    def getDefaultBeamline():
        hostname = platform.node()
        for beamline in PublishProcessedDataWidget.KNOW_BEAMLINES:
            if beamline in hostname:
                return beamline

        return None

    def set_auto_update(self, update: bool):
        self._beamlinePLB.setChecked(not update)
        self._proposalPLB.setChecked(not update)
        self._datasetPLB.setChecked(not update)

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        return {
            "beamline_auto_update": not self._beamlinePLB.isChecked(),
            "dataset_auto_update": not self._datasetPLB.isChecked(),
            "proposal_auto_update": not self._proposalPLB.isChecked(),
            "beamline": self._beamlineCB.currentText(),
            "dataset": self._datasetQLE.text(),
            "proposal": self._proposalQLE.text(),
        }

    def setConfiguration(self, configuration: dict):
        with block_signals(self):
            # handle beamline
            beamline = configuration.get("beamline", None)
            if beamline is not None:
                self._beamlineCB.setCurrentText(beamline)
            beamline_auto_update = configuration.get("beamline_auto_update", None)
            if beamline_auto_update is not None:
                self._beamlinePLB.setChecked(not beamline_auto_update)

            # handle dataset
            dataset = configuration.get("dataset", None)
            if dataset is not None:
                self._datasetQLE.setText(dataset)
            dataset_auto_update = configuration.get("dataset_auto_update", None)
            if dataset_auto_update is not None:
                self._datasetPLB.setChecked(not dataset_auto_update)

            # handle proposal
            proposal = configuration.get("proposal", None)
            if proposal is not None:
                self._proposalQLE.setText(proposal)
            proposal_auto_update = configuration.get("proposal_auto_update", None)
            if proposal_auto_update is not None:
                self._proposalPLB.setChecked(not proposal_auto_update)

        self._configChanged()

    def setScan(self, scan: TomwerScanBase):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase}, {type(scan)} provided instead"
            )

        new_config = {}
        if not self._proposalPLB.isChecked():
            new_config["proposal"] = scan.get_proposal_name()
        if not self._beamlinePLB.isChecked():
            new_config["beamline"] = scan.instrument_name
        if not self._datasetPLB.isChecked():
            new_config["dataset"] = scan.sample_name

        self.setConfiguration(new_config)



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "03/05/2019"


import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.saaxis import SAAxisTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.saaxis import QSAAxisParams

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class SAAxisProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, saaxis_params, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)
        assert saaxis_params is not None
        self._dry_run = False
        self._process_fct = None

    def patch_processing(self, process_fct):
        self._computationThread.patch_processing(process_fct)

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(f"start saaxis {data}")
        assert isinstance(data, TomwerScanBase)
        if data.axis_params is None:
            data.axis_params = QAxisRP()
        if data.saaxis_params is None:
            data.saaxis_params = QSAAxisParams()
        self._data_currently_computed = data
        saaxis_params = QSAAxisParams.from_dict(configuration)
        saaxis_params.frame_width = data.dim_1
        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = f"low memory, skip saaxis calculation {data.path}"
            ProcessManager().notify_dataset_state(
                dataset=data, process=self._process_id, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            data.axis_params.set_relative_value(None)
            if callback is not None:
                callback()
            self.scan_ready(scan=data)
        else:
            self._data_currently_computed = data
            self._computationThread.init(data=data, saaxis_params=saaxis_params)
            # need to manage connect before starting it because
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.start()

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        assert isinstance(data, TomwerScanBase)
        FIFO._end_computation(
            self, data=data, future_tomo_obj=future_tomo_obj, callback=callback
        )

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of the axis position
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self.center_of_rotation = None
        self._dry_run = False
        self._scan = None
        self._saaxis_params = None
        self._patch_process_fct = None
        """function pointer to know which function to call for the axis
        calculation"""
        self._current_processing = None
        # processing currently runned. As we want to be able to cancel processing we need to keep a pointer on the
        # 'activate' nabu processing to stop process if requested
        self._task = None
        # ewoks task for reconstruction

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def patch_processing(self, process_fct):
        self._patch_process_fct = process_fct

    def init(self, data, saaxis_params: dict):
        self._scan = data
        self._saaxis_params = saaxis_params

    def run(self):
        self.sigComputationStarted.emit()
        if self._patch_process_fct:
            scores = {}
            for cor in self._saaxis_params.cors:
                scores[cor] = self._patch_process_fct(cor)
            self._scan.saaxis_params.scores = scores
            SAAxisTask.autofocus(scan=self._scan)
            self.center_of_rotation = self._scan.saaxis_params.autofocus
        else:
            self._task = SAAxisTask(
                process_id=self.process_id,
                inputs={
                    "dump_process": False,
                    "data": self._scan,
                    "dry_run": self._dry_run,
                    "sa_axis_params": self._saaxis_params.to_dict(),
                    "serialize_output_data": False,
                },
            )

            self._task.run()
            self.center_of_rotation = self._task.outputs.best_cor

    def cancel(self):
        if self._task is not None:
            self._task.cancel()
        self.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import signal
import sys
import weakref

import silx
from silx.gui import qt

from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.visualization.sinogramviewer import SinogramViewer as _SinogramViewer

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class SinogramViewer(_SinogramViewer):
    def setScan(self, scan):
        if self._scan is None or self._scan() != scan:
            self._scan = weakref.ref(scan)
            self._options.setScan(scan)


def getinputinfo():
    return "tomwer sinogram-viewer [scan_path --entry entry --line line --subsampling subsampling]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.closeAllWindows()  # needed because get a waiting thread behind
    qt.QApplication.quit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="path to the acquisition (folder for EDF, master file for HDF5)",
    )
    parser.add_argument(
        "--entry",
        default=None,
        help="For HDF5 acquisition you should provide an entry if several"
        "acquisitions are contained in the master file",
    )
    parser.add_argument(
        "--line",
        default=None,
        help="line to extract from radios to create the sinogram. Take the "
        "middle of projections by default",
    )
    parser.add_argument(
        "--subsampling",
        default=1,
        help="You can define a subsampling to generate the sinogram in order"
        "to speed up creation",
    )
    parser.add_argument(
        "--use-opengl-plot",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(argv[1:])

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    scan = ScanFactory.create_scan_object(options.scan_path, entry=options.entry)
    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)

    if options.subsampling < 1:
        raise ValueError("subsampling should be at least 1")

    if options.line is None:
        options.line = scan.dim_2 // 2

    if options.line < 0:
        raise ValueError("Line value should be at least 0")
    if options.line > scan.dim_2:
        raise ValueError(
            f"Line value is outside frame dimension. Should be at most {scan.dim_2}"
        )

    widget = SinogramViewer(parent=None)
    widget.setLine(options.line)
    widget.setSubsampling(options.subsampling)
    widget.setScan(scan)
    widget.setWindowTitle("Sinogram viewer")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)
